/**
A complete search method for the relaxed traveling tournament problem.

Copyright (C) 2013, Filipe Brandao
Faculdade de Ciencias, Universidade do Porto
Porto, Portugal. All rights reserved. E-mail: <fdabrandao@dcc.fc.up.pt>.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#include<stdio.h>
#include<limits.h>
#include<string.h>
#include<math.h>
#include<time.h>
#include<assert.h>
#include<stdlib.h>

#define true (1)
#define false (0)
#define swap(x,y) { int t; t = x; x = y; y = t; }

#define INT_INF (INT_MAX>>1)

#define MAXSTR (100)
#define VERBOSE (0)
#define MAX_N (8)
#define MAX_K (10)
#define MAX_CHOME (3)                // limit on the length of home stands
#define MAX_CAWAY (3)                // limit on the length of road trips
#define MAX_NR (2*(MAX_N-1))         // maximum number of rounds
#define MAX_NRK (MAX_NR+MAX_K)       // maximum number of rounds + byes
#define GTEAM(x) (abs(x)-1)          // game -> opponent's index
#define GTYPE(x) ((x) > 0 ? 1 : -1)  // game -> 1: home, -1: away

//Additional constraints:
#define FILL_LTSLOT (0)              // force every team to play in the last time slot
#define MAX_CHOME_BYES (K)           // limit on the number of consecutive home byes
#define MAX_CAWAY_BYES (K)           // limit on the number of consecutive away byes

#define OPT_ITR 100                  // number of hill-climbing iterations 

typedef char bool;

int N;         // number of teams
int K;         // number of byes
int NR;        // number of rounds w/o byes
int NRK;       // number of rounds
int ITR_COUNT; // number of iterations
int CUR_UB;    // current UB

int dist[MAX_N][MAX_N]; // distance matrix


// schedules:
//    values:
//       0: bye,  > 0: home game, < 0: away game
//    GTEAM(schedule[round][team]) -> opponent's index
//    GTYPE(schedule[round][team]) -> 1: home, -1: away
int SCH[MAX_NR+MAX_K][MAX_N];       // current schedule
int BEST_SCH[MAX_NR+MAX_K][MAX_N];  // best schedule (B&B)
int LAST_SCH[MAX_NR+MAX_K][MAX_N];  // last schedule (B&B)
int HBEST_SCH[MAX_NR+MAX_K][MAX_N]; // best schedule (MH)

bool G[MAX_N][MAX_N];         // true if (t1,t2) has already been scheduled
int GAME_ROUND[MAX_N][MAX_N]; // maps each game (t1,t2) to a round

int RGS[MAX_N];     // remaining game slots
int CUR_LOC[MAX_N]; // current location
int LAST_OP[MAX_N]; // last opponent

int BYES[MAX_N]; // number of byes used

int H[MAX_N];    // current number of consecutive home games
int NH[MAX_N];   // current number of consecutive away games

int CB[MAX_N];   // current number of consecutive byes

char instance[MAXSTR];
time_t T_START;

char names[MAX_N][MAXSTR]; // team names

/************************
*                       *
*   Input/Output Code   *
*                       *
************************/

// read: loads a RTTP instance
void read(const char *file_instance, const char *file_names) {  
  FILE *f;
  int cnt;
  if(file_names != NULL){
    f = fopen(file_names, "r");
    if(f == NULL) {
      printf("failed to open file '%s'\n", file_names);
      exit(1);
    }
    cnt = 0;
    char buf[MAXSTR];
    while(fscanf(f, "%s", buf) != EOF)
      strcpy((char*)&names[cnt++], buf);
    fclose(f);
  } else {
    int i;
    for(i = 0; i < MAX_N; i++)
      sprintf((char*)&names[i], "C%.2d", i+1);
    cnt = MAX_N;
  }
  f = fopen(file_instance, "r");
  if(f == NULL) {
    printf("failed to open file '%s'\n", file_instance);
    exit(1);
  }
  N = 0;
  while(fscanf(f, "%*d")==0) N++;
  N = (int)sqrt(N);
  assert(N <= cnt);
  assert(N <= MAX_N);
  NR = 2*(N-1);
  NRK = NR+K;
  fseek(f, 0, SEEK_SET);
  int r, c;
  for(r = 0; r < N; r++) {
    for(c = 0; c < N; c++) {
      assert(fscanf(f, "%d", &dist[r][c])==1);
    }
  }
  fclose(f);
}

// print_schedule_trick: prints the schedule SCH in trick's format
void print_schedule_trick(int sch[MAX_NR+MAX_K][MAX_N], FILE *out) {
  int r, t;
  for(t = 0; t < N; t++) {
    fprintf(out, "%-6s", names[t]);
  }
  fputc('\n', out);
  for(t = 0; t < N; t++) {
    fprintf(out, "---   ");
  }
  fputc('\n', out);
  char name[MAXSTR] = "@";
  for(r = 0; r < NRK; r++) {
    for(t = 0; t < N; t++) {
      int t2 = sch[r][t];
      if(t2 == 0) {
        fprintf(out, "***   ");
      } else if(t2 > 0) {
        fprintf(out, "%-6s",names[GTEAM(t2)]);
      } else if(t2 == INT_MIN) {
        fprintf(out, "xxx   ");
      } else {
        name[1] = 0;
        strcat(name, names[GTEAM(t2)]);
        fprintf(out, "%-6s",name);
      }
    }
    fputc('\n', out);
  }
}

// save: saves the solution SCH to a file
bool is_valid();
void save(int sch[MAX_NR+MAX_K][MAX_N], const char *filename) {
  assert(is_valid() == true);
  FILE *out = fopen(filename, "w");
  print_schedule_trick(sch, out);
  fprintf(out, "Instance: %s, N = %d, K = %d\n", instance, N, K);
  fprintf(out, "Cost: %d\n", CUR_UB);
  fprintf(out, "Format: Trick\n");
  fprintf(out, "Found by: Brandão and Pedroso\n");
  fclose(out);
}

/******************
*                 *
*   Common code   *
*                 *
******************/

// clear: clears auxiliary data structures
void clear() {
  int r, t, i;
  for(r = 0; r < NRK; r++) {
    for(t = 0; t < N; t++) {
      SCH[r][t] = INT_MIN;
    }
  }
  for(i = 0; i < N; i++) {
    BYES[i] = K;
    CUR_LOC[i] = i;
    LAST_OP[i] = i;
    RGS[i] = NR;
    CB[i] = 0;
  }
  memset(G, false, sizeof(G));
  memset(H, 0, sizeof(H));
  memset(NH, 0, sizeof(NH));
}

// get_cost(R): computes the cost (until round R) of the current solution SCH
int get_cost_r(int R) {
  int r, t;
  int cost = 0;
  static int CL[MAX_N];
  for(t = 0; t < N; t++) {
    CL[t] = t;
  }
  for(r = 0; r < R; r++) {
    for(t = 0; t < N; t++) {
      int a = CL[t];
      int b = SCH[r][t];
      if(b != 0) {
        if(b > 0) b = t;
        if(b < 0) b = GTEAM(b);
        cost += dist[a][b];
        CL[t] = b;
      }
    }
  }
  return cost;
}

// get_cost: computes the cost of the current solution SCH
int get_cost() {
  int r, t;
  int cost = 0;
  static int CL[MAX_N];
  for(t = 0; t < N; t++) {
    CL[t] = t;
  }
  for(r = 0; r < NRK; r++) {
    for(t = 0; t < N; t++) {
      int a = CL[t];
      int b = SCH[r][t];
      if(b != 0) {
        if(b > 0) b = t;
        if(b < 0) b = GTEAM(b);
        cost += dist[a][b];
        CL[t] = b;
      }
    }
  }
  for(t = 0; t < N; t++) {
    int a = CL[t];
    cost += dist[a][t];
  }
  return cost;
}

// is_valid: checks if every constraint is respected in the current solution SCH
bool is_valid() {
  int r, t, t1, t2;
  for(r = 0; r < NRK; r++) {
    for(t1 = 0; t1 < N; t1++) {
      if(SCH[r][t1] == 0) continue;
      t2 = GTEAM(SCH[r][t1]);
      if(GTEAM(SCH[r][t2]) != t1) return false;
      if(SCH[r][t1]*SCH[r][t2] >= 0) return false;
    }
  }
  static bool G[MAX_N][MAX_N];
  memset(G, false, sizeof(G));
  for(t = 0; t < N; t++) {
    int h = 0;
    int nh = 0;
    int cb = 0;
    int lop = -1;
    if(FILL_LTSLOT) {
      if(SCH[NRK-1][t] == 0) return false;
    }
    for(r = 0; r < NRK; r++) {
      const int sch_rt = SCH[r][t];
      if(sch_rt != 0) {
        if(sch_rt > 0) G[t][GTEAM(sch_rt)] = true;  
        if(r > 0 && GTEAM(sch_rt) == lop) return false;
        lop = GTEAM(sch_rt);
      }
      if(sch_rt > 0) { // home
        if(h >= MAX_CHOME) return false;
        h++;
        nh = 0;
        cb = 0;
      } else if(sch_rt < 0) { // away
        if(nh >= MAX_CAWAY) return false;
        nh++;
        h = 0;
        cb = 0;
      } else { // bye
        if(nh > 0){ // away bye
          if(cb >= MAX_CAWAY_BYES) return false;
        } else { // home bye
          if(cb >= MAX_CHOME_BYES) return false;        
        }
        cb++;
      }
    }
  }
  int i, k;  
  for(i = 0; i < N; i++) {
    for(k = 0; k < N; k++) {
      if(i != k && !G[i][k]) {
        return false;
      }
    }
  }
  return true;
}

/*********************
*                    *
*   Metaheuristics   *
*                    *
*********************/

// swap_homes(...): swaps the two games between two teams
void swap_homes(int r1, int r2, int t1, int t2) {
  SCH[r1][t1] *= -1;
  SCH[r1][t2] *= -1;
  SCH[r2][t1] *= -1;
  SCH[r2][t2] *= -1;
}

// swap_roles: swaps home/away roles in every game
void swap_roles() {
  int r, t;
  for(r = 0; r < NRK; r++) {
    for(t = 0; t < N; t++) {
      SCH[r][t] *= -1;
    }
  }
}

// swap_rounds: swaps two rounds
void swap_rounds(int r1, int r2) {
  int i;
  for(i = 0; i < N; i++) {
    swap(SCH[r1][i],SCH[r2][i]);
  }
}

// swap_teams: swaps the schedule of two teams
void swap_teams(int a, int b) {
  int r, t1, t2, s;
  for(r = 0; r < NRK; r++) {
    for(t1 = 0; t1 < N; t1++) {
      if(!SCH[r][t1]) continue;
      s = GTYPE(SCH[r][t1]);
      t2 = GTEAM(SCH[r][t1]);
      if(t2 == a) {
        SCH[r][t1] = s*(b+1);
      } else if(t2 == b) {
        SCH[r][t1] = s*(a+1);
      }
    }
    swap(SCH[r][a],SCH[r][b]);
  }
}

// shuffle: applies random perturbations to SCH
void shuffle(int sr, int st, int sh) {
  while(sr > 0 || st > 0 || sh > 0) {
    while(sr-- > 0) {
      int r1 = rand()%NRK;
      int r2 = rand()%NRK;
      if(r2 == r1) r2 = (r2+1)%NRK;
      swap_rounds(r1, r2);
      if(!is_valid()) swap_rounds(r1, r2);
    }
    while(st-- > 0) {
      int t1 = rand()%N;
      int t2 = rand()%N;
      if(t2 == t1) t2 = (t2+1)%N;
      swap_teams(t1, t2);
    }
    while(sh-- > 0) {
      int t1 = rand()%N;
      int t2 = rand()%N;
      if(t2 == t1) t2 = (t2+1)%N;
      int r1 = GAME_ROUND[t1][t2];
      int r2 = GAME_ROUND[t2][t1];
      swap_homes(r1, r2, t1, t2);
      if(!is_valid()) swap_homes(r1, r2, t1, t2);
    }
  }
}

// search_neighborhood: hill climbing iteration
int search_neighborhood(int cur_best) {
  assert(is_valid());
  int r, r1, r2, t1, t2;
  int min_cost = cur_best;
  for(r = 0; r < NRK; r++) {
    for(t1 = 0; t1 < N; t1++) {
      t2 = SCH[r][t1];
      if(t2 > 0) {
        GAME_ROUND[t1][t2] = r;
      }
    }
  }
  memcpy(HBEST_SCH, SCH, sizeof(SCH));

  swap_roles();
  int cost = get_cost();
  if(cost < min_cost && is_valid()) {
    min_cost = cost;
    memcpy(HBEST_SCH, SCH, sizeof(SCH));
  }
  swap_roles();

  //swap homes
  for(t1 = 0; t1 < N; t1++) {
    for(t2 = t1+1; t2 < N; t2++) {
      r1 = GAME_ROUND[t1][t2];
      r2 = GAME_ROUND[t2][t1];
      swap_homes(r1, r2, t1, t2);
      if(is_valid()) {
        int cost = get_cost();
        if(cost < min_cost) {
          min_cost = cost;
          memcpy(HBEST_SCH, SCH, sizeof(SCH));
        }
      }
      swap_homes(r1, r2, t1, t2);
    }
  }

  //swap opponents
  for(t1 = 0; t1 < N; t1++) {
    for(t2 = t1+1; t2 < N; t2++) {
      swap_teams(t1,t2);
      //assert(is_valid());
      int cost = get_cost();
      if(cost < min_cost) {
        min_cost = cost;
        memcpy(HBEST_SCH, SCH, sizeof(SCH));
      }
      swap_teams(t1,t2);
    }
  }

  //swap rounds
  for(r1 = 0; r1 < NRK; r1++) {
    for(r2 = r1+1; r2 < NRK; r2++) {
      swap_rounds(r1,r2);
      if(is_valid()) {
        int cost = get_cost();
        if(cost < min_cost) {
          min_cost = cost;
          memcpy(HBEST_SCH, SCH, sizeof(SCH));
        }
      }
      swap_rounds(r1,r2);
    }
  }

  memcpy(SCH, HBEST_SCH, sizeof(SCH));
  return min_cost;
}

// hill_climbing: tries to improve SCH using hill climbing
int hill_climbing() {
  int cost = INT_INF;
  if(VERBOSE) printf("%d", get_cost());
  while(true) {
    int t = search_neighborhood(cost);
    if(t >= cost) break;
    else {
      cost = t;
      if(VERBOSE) putchar('.');
    }
  }
  if(VERBOSE) {
    printf("%d;\n", cost);
    fflush(stdout);
  }
  return cost;
}

// optimize: tries to improve SCH using metaheuristics
void optimize(int count) {
  int i;
  static char name1[MAXSTR];
  static char name2[MAXSTR];  
  static char S[sizeof(SCH)];
  memcpy(S, SCH, sizeof(SCH));
  for(i = 0; i < count; i++) {
    int cost = hill_climbing(); // call the hill climbing procedure
    if(cost < CUR_UB) {
      CUR_UB = cost;
      memcpy(BEST_SCH, HBEST_SCH, sizeof(SCH));
      printf("**> %d (%lds)\n", CUR_UB, time(0)-T_START);
      sprintf(name1,"solutions/rttp_n%d_k%d_h%d.txt", N, K, CUR_UB);
      sprintf(name2,"rttp_n%d_k%d.txt", N, K);
      save(BEST_SCH, "out.txt");
      save(BEST_SCH, name1);
      save(BEST_SCH, name2);
    }
    shuffle(10, 10, 10); // apply random perturbations
  }
  memcpy(SCH, S, sizeof(SCH));
}

/******************************
*                             *
*   Independent Lower Bound   *
*                             *
*******************************/

// hash: function to compute the DP state key
#define DP_SIZE 4096 // [2^N+2] 2^(9+2+1) N=10 16MB
static int dp[10][10][10][DP_SIZE];
int hash(int t, int h, int nh) {
  int i, k = 0;
  for(i = 0; i < N; i++) {
    if(i==t) continue;
    k <<= 1;
    k |= G[i][t];
  }
  k <<= 2;
  k |= h|nh;
  k <<= 1;
  k |= (nh!=0);
  assert(h <= 3);
  assert(nh <= 3);
  assert(!h || !nh);
  return k;
}

// go_ilb: recursive function to compute the ILB using DP
int ilb_cur_loc, ilb_h, ilb_nh, ilb_home_games;
int go_ilb(int t, int g) {
  if(g == 0) {
    return dist[ilb_cur_loc][t];
  } else {
    int i;
    int key = hash(t, ilb_h, ilb_nh);
    assert(key>=0 && key<DP_SIZE);
    if(dp[t][ilb_cur_loc][ilb_home_games][key]) {
      return dp[t][ilb_cur_loc][ilb_home_games][key]-1;
    } else {
      static int dp_count = 0;
      dp_count++;
    }
    int best = INT_INF;
    int d = dist[ilb_cur_loc][t];
    int t_ilb_h = ilb_h;
    int t_ilb_nh = ilb_nh;
    int t_ilb_cur_loc = ilb_cur_loc;
    if(ilb_h < MAX_CHOME) {
      ilb_h++;
      ilb_nh = 0;
      ilb_home_games--;
      for(i = 0; i < N; i++) {
        if(i==t) continue;
        if(!G[t][i]) { //@t
          G[t][i] = true;
          ilb_cur_loc = t;
          int tmp = go_ilb(t, g-1);
          if(tmp < INT_INF) {
            tmp += d;
            if(tmp < best) best = tmp;
          }
          G[t][i] = false;
        }
      }
      ilb_home_games++;
      ilb_h = t_ilb_h;
      ilb_nh = t_ilb_nh;
      ilb_cur_loc = t_ilb_cur_loc;
    }
    if(ilb_nh < MAX_CAWAY) {
      ilb_h = 0;
      ilb_nh++;
      for(i = 0; i < N; i++) {
        if(i==t) continue;
        if(!G[i][t]) { //@i
          d = dist[t_ilb_cur_loc][i];
          G[i][t] = true;
          ilb_cur_loc = i;
          int tmp = go_ilb(t, g-1);
          if(tmp < INT_INF) {
            tmp += d;
            if(tmp < best) best = tmp;
          }
          G[i][t] = false;
        }
      }
      ilb_h = t_ilb_h;
      ilb_nh = t_ilb_nh;
      ilb_cur_loc = t_ilb_cur_loc;
    }
    assert(best >= 0);
    assert(best <= INT_INF);
    dp[t][ilb_cur_loc][ilb_home_games][key] = best+1;
    return best;
  }
}

// get_ilb_t(t): function to compute the ILB for each team t
int get_ilb_t(int t) {
  int k;
  ilb_h = H[t];
  ilb_nh = NH[t];
  ilb_home_games = 0;
  for(k = 0; k < N; k++) {
    if(t!=k) ilb_home_games += !G[t][k];
  }
  ilb_cur_loc = CUR_LOC[t];
  return go_ilb(t, RGS[t]);
}

// get_ilb: funcion to compute the ILB
int get_ilb() {
  int i;
  int s = 0;
  for(i = 0; i < N; i++) {
    int v = get_ilb_t(i);
    if(v == INT_INF) return INT_INF;
    s += v;
  }
  return s;
}

/***********************
*                      *
*   Branch-and-Bound   *
*                      *
***********************/

//go: recursive depth-first branch-and-bound procedure
int _cost = 0;
bool go(int r, int p) {
  if(_cost >= CUR_UB) {
    return false;
  }

  if(p == N) {
    //assert(get_cost_r(r+1) == _cost);
    if(r == NRK-1) { // complete schedule
      memcpy(LAST_SCH, SCH, sizeof(SCH));
      int t;
      int cost = _cost;
      for(t = 0; t < N; t++) {
        cost += dist[CUR_LOC[t]][t];
      }
      if(cost < CUR_UB) {
        memcpy(BEST_SCH, SCH, sizeof(SCH));
        CUR_UB = cost;
        static char name1[MAXSTR], name2[MAXSTR];
        sprintf(name1,"solutions/rttp_n%d_k%d_b%d.txt", N, K, CUR_UB);
        sprintf(name2,"rttp_n%d_k%d.txt", N, K);
        save(BEST_SCH, "out.txt");
        save(BEST_SCH, name1);
        save(BEST_SCH, name2);
        printf("*> %d (%d, %lds)\n", CUR_UB, ITR_COUNT, time(0)-T_START);
      }
      assert(cost == get_cost());
      assert(is_valid());
      optimize(OPT_ITR);
      ITR_COUNT++;
      return false;
    } else if(r >= 0) {
      //assert(_cost==get_cost(r+1));
      int ilb = get_ilb();
      if(_cost + ilb >= CUR_UB) return false;
    }
    return go(r+1, 0);
  }

  if(SCH[r][p] != INT_MIN) { // game already schduled
    RGS[p]--;
    if(go(r, p+1)) return true;
    RGS[p]++;
    return false;
  }

  if(RGS[p] > 0) { // if there are available game slots
    int i, k;
    RGS[p]--;    
    int u = p;
    int clu = CUR_LOC[u];
    int lopu = LAST_OP[u]; 
    for(i = p+1; i < N && _cost < CUR_UB; i++) {
      if(LAST_OP[p] == i || LAST_OP[i] == p || SCH[r][i] != INT_MIN) continue;
      
      int v = i;
      int clv = CUR_LOC[v];
      int lopv = LAST_OP[v];
    
      for(k = 0; k < 2; k++) {
        if(!G[u][v] && H[u] < MAX_CHOME && NH[v] < MAX_CAWAY) { //@u
          int d = dist[clu][u]+dist[clv][u];
          if(_cost+d < CUR_UB){
              // setup solution
              int cb_u = CB[u], cb_v = CB[v];
              int nh_u = NH[u], h_v = H[v];
              CB[u] = CB[v] = 0;
              H[u]++;
              NH[u] = 0;
              H[v] = 0;
              NH[v]++;
              LAST_OP[u] = v;
              LAST_OP[v] = u;
              SCH[r][u] = v+1;
              SCH[r][v] = -u-1;
              _cost += d;
              G[u][v] = true;
              CUR_LOC[u] = CUR_LOC[v] = u;
              // recursive call
              if(go(r, p+1))
                return true; 
              // restore solution
              G[u][v] = false;
              _cost -= d;
              H[u]--;
              NH[u] = nh_u;
              H[v] = h_v;
              NH[v]--;
              CUR_LOC[u] = clu;
              CUR_LOC[v] = clv;
              LAST_OP[u] = lopu;
              LAST_OP[v] = lopv;
              CB[u] = cb_u; CB[v] = cb_v;
          }
        }
        swap(u, v);
        swap(clu, clv);
        swap(lopu, lopv);
      }
      SCH[r][u] = INT_MIN;
      SCH[r][v] = INT_MIN;
    }
    RGS[p]++;
  }
  
  if(FILL_LTSLOT && r == NRK-1) return false;

  if(RGS[p] < NRK-r) { //byes
    if(CUR_LOC[p] == p) {
      if(CB[p] >= MAX_CHOME_BYES) return false;
    } else {
      if(CB[p] >= MAX_CAWAY_BYES) return false;   
    }
    CB[p]++;    
    SCH[r][p] = 0;
    if(go(r, p+1)) return true;
    SCH[r][p] = INT_MIN;
    CB[p]--;   
  }
  return false;
}

//solve: calls the B&B function
void solve() {
  ITR_COUNT = 0;
  T_START = time(0);
  clear();
  printf("ILB: %d\n", get_ilb());
  _cost = 0;
  go(0,0);
  if(CUR_UB == INT_INF)
    printf(">> unfeasible!\n");
  else
    printf(">> Cost: %d (%d, %ds)\n", CUR_UB, ITR_COUNT, (int)(time(0)-T_START));    
}

int main(int argc, char *argv[]) {
  setvbuf(stdout, NULL, _IONBF, 0);
  if(argc >= 2) 
    strcpy(instance, argv[1]);
  else 
    strcpy(instance, "instances/data6.txt");
  printf("%s\n", instance);

  if(argc >= 3) 
    K = atoi(argv[2]);
  else 
    K = 1;
  printf("K: %d\n", K);

  if(argc >= 4)
    read(instance, argv[3]);
  else
    read(instance, "instances/names_nl.txt");  

  if(argc >= 5){
    srand(atoi(argv[4]));
  } else {
    srand(1234);  
  }

  if(argc >= 6) {
    CUR_UB = atoi(argv[5]);
    printf("UB: %d\n", CUR_UB);
  } else {
    CUR_UB = INT_INF;
    printf("UB: inf\n");
  }
  solve();
  return 0;
}


